"""Tuning API for embedding, reranking, and LLM"""


class AutoTuner(object):
    """Given the data and model, assign the tuning type, then fine-tune for embedding, ranking and llm"""

    def __init__(self):
        pass

    def run(self):
        pass
